/***********************************************************
 * $Id: HiLoIdGenerator.java 794 2013-02-19 15:46:41Z util $
 * 
 * SQL/DAO utilities of clazzes.org
 * http://www.clazzes.org
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 * 
 ***********************************************************/

package org.clazzes.util.sql.dao;

import java.sql.Connection;
import java.sql.ResultSet;
import java.sql.SQLException;
import java.sql.Statement;

import javax.sql.DataSource;

import org.clazzes.util.aop.DAOException;
import org.clazzes.util.aop.jdbc.JdbcStatementAction;
import org.clazzes.util.sql.helper.JDBCHelper;
import org.clazzes.util.sql.helper.JDBCTransaction;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

/**
 * <p>A Hi-Lo ID generator, which generates IDs in block with
 * a configurable bock size.</p>
 * 
 * <p>Be sure to call {@link #initialize()} after setting a valid datasource
 * via {@link #setDataSource(DataSource)} and actually creating your generator
 * database table.</p>
 * 
 * <p>The code is placed in this abstract class, since some details have to be
 * done in slightly different way when dealing with a Hibernate-controlled database.
 * This means, usually the subclass HiLoIdGenerator should be used.  Only when
 * supporting a database Hibernate controls, the other subclass HibernateHiLoIdGenerator
 * should be used, as hibernate stores values divided by the block size in the
 * id table, whereas HiLoIdGenerator stores values without division. </p>
 */
public abstract class AbstractHiLoIdGenerator implements IdGenerator {
    
    private static final Logger log = LoggerFactory.getLogger(HiLoIdGenerator.class);

    public static final String DEFAULT_ID_TABLE_NAME = "ID_GENERATOR";
    public static final String DEFAULT_NEXT_ID_COL_NAME = "NEXT_ID";
    
    private String idTableName;
    private String nextIdColumnName;
    private long blockSize;
    private DataSource dataSource;
    
    private Long hiValue;
    private long loValue;
    
    /** Returns the very first hi value the generator writes into the database (if none was written before).
     * @return the very first hi value the generator writes into the database (if none was written before).
     */
    protected abstract Long getInitialHiValue();
    
    /** Returns the step size, by which the hi value (as stored in the database) is incremented when a new
     *  block has to be started.
     * @return step size as described
     */
    protected abstract Long getHiStepSize();
    
    /** Returns the current base id, to which the lo value is added.  It depends on the hi value stored in the
     *  database, but there might be some factor in between that is encapsulated in this function.
     * @return the current base id as described
     */
    protected abstract Long getBaseId();    

    public AbstractHiLoIdGenerator() {
        this.idTableName = DEFAULT_ID_TABLE_NAME;
        this.nextIdColumnName = DEFAULT_NEXT_ID_COL_NAME;
        this.blockSize = 100L;
    }
    
    protected Long runInSerializableTransaction(JdbcStatementAction<Long> action) throws SQLException {
        
        try (JDBCTransaction txn = new JDBCTransaction(this.dataSource,Connection.TRANSACTION_SERIALIZABLE)) {
            
            Statement statement = null;
            
            try {
                
                statement = txn.getConnection().createStatement();
                Long ret = action.perform(statement);
                txn.commit();
                return ret;
            }
            catch (Throwable e) {
                
                if (statement != null) {
                    try {
                        statement.close();
                    }
                    catch (SQLException e2) {
                        log.warn("Error closing SQLStatement on Id table ["+this.idTableName+"]",e2);
                    }
                }
                
                throw new DAOException("Error operating on Id table ["+this.idTableName+"]",e);
            }
        }
    }
    
    protected Long fetchNextId(Statement statement) throws SQLException {
        
        ResultSet rs = statement.executeQuery("select "+
                AbstractHiLoIdGenerator.this.nextIdColumnName
                +" from "+ AbstractHiLoIdGenerator.this.idTableName);
        
        Long nextId = null;
        
        try {
            while (rs.next()) {
                
                if (nextId == null) {
                    nextId = JDBCHelper.getLong(rs,1);
                }
                else {
                    throw new DAOException("Id table ["+AbstractHiLoIdGenerator.this.idTableName+"] contains multiple rows.");
                }
            }
        }
        finally {
            rs.close();
        }
        
        return nextId;
    }
    

    
    /**
     * An initialization function, which check the database table for a single row
     * with the next high value.
     * @throws SQLException
     */
    public void initialize() throws SQLException {
        
        this.hiValue =
                this.runInSerializableTransaction(new JdbcStatementAction<Long>() {
                    
                    @Override
                    public Long perform(Statement statement) throws Exception {
                        
                        Long nextId = AbstractHiLoIdGenerator.this.fetchNextId(statement);
                        
                        if (nextId == null) {
                            
                            nextId = AbstractHiLoIdGenerator.this.getInitialHiValue();
                            
                            log.info("Id table [{}] contains no rows, inserting initial value [{}].",
                                    AbstractHiLoIdGenerator.this.idTableName,nextId);
                            
                            statement.executeUpdate("insert into "+AbstractHiLoIdGenerator.this.idTableName+
                                    " (" + AbstractHiLoIdGenerator.this.nextIdColumnName+") values ("+nextId+")");
                            
                        }
                        
                        return nextId;  
                    }
                });
        
        if (log.isDebugEnabled()) {
            log.debug("Last known high value was [{}].",this.hiValue);
        }

        this.loValue = 0;
    }
    
    /* (non-Javadoc)
     * @see org.clazzes.util.sql.dao.IdGenerator#generateNext()
     */
    @Override
    public synchronized Long generateNext() {
        
        if (this.hiValue == null || this.loValue >= 0) {
        
            try {
            
                if (log.isDebugEnabled()) {
                    log.debug("Generating new high value as successor to [{}].",this.hiValue);
                }
                
                this.hiValue = 
                        this.runInSerializableTransaction(new JdbcStatementAction<Long>() {
                            
                            @Override
                            public Long perform(Statement statement) throws Exception {
                                
                                statement.executeUpdate("update "+ AbstractHiLoIdGenerator.this.idTableName+
                                        " set "+ AbstractHiLoIdGenerator.this.nextIdColumnName+"="+
                                        AbstractHiLoIdGenerator.this.nextIdColumnName+"+" + AbstractHiLoIdGenerator.this.getHiStepSize());

                                
                                Long nextId = AbstractHiLoIdGenerator.this.fetchNextId(statement);
                                
                                if (nextId == null) {
                                    throw new DAOException("Id table [" + AbstractHiLoIdGenerator.this.idTableName+"] contains no rows, did you call initialize()?");
                                }
                                return nextId;  
                            }
                        });
                
                if (log.isDebugEnabled()) {
                    log.debug("New high value is [{}].",this.hiValue);
                }

                this.loValue = -this.blockSize;
                
            } catch (SQLException e) {
                throw new DAOException("Error generating new High value",e);
            }
        }   
        
        Long ret = Long.valueOf(this.getBaseId().longValue() + this.loValue++);
        
        if (log.isDebugEnabled()) {
            log.debug("Generated new ID [{}].",ret);
        }
        
        return ret;
    }
    
    /**
     *  Getter to the hi value as stored in the corresponding database table, in order to make
     *  it available for test cases.
     * @return the hi value this generator currently operates on.
     */
    public Long getHiValue() {
        return this.hiValue;
    }

    /**
     * @return The name of the ID table.
     */
    public String getIdTableName() {
        return this.idTableName;
    }

    /**
     * @param idTableName THe name of the ID table. The default value
     *         is <code>"ID_GENERATOR"</code> resp. @see DEFAULT_ID_TABLE_NAME.
     */
    public void setIdTableName(String idTableName) {
        this.idTableName = idTableName;
    }

    /**
     * @return The name of the column, where the next hi value is stored.
     */
    public String getNextIdColumnName() {
        return this.nextIdColumnName;
    }

    /**
     * @param nextIdColumnName The name of the column,
     *        where the next hi value is stored. The default value is
     *         <code>"NEXT_ID"</code> resp. @see DEFAULT_NEXT_ID_COL_NAME
     */
    public void setNextIdColumnName(String nextIdColumnName) {
        this.nextIdColumnName = nextIdColumnName;
    }

    /**
     * @return The block size for high values.
     */
    public long getBlockSize() {
        return this.blockSize;
    }

    /**
     * @param blockSize The block size to set. The default value is
     *         <code>100</code>.
     */
    public void setBlockSize(long blockSize) {
        this.blockSize = blockSize;
    }

    /**
     * @return The datasource used to pull new connections.
     */
    public DataSource getDataSource() {
        return this.dataSource;
    }

    /**
     * @param dataSource The datasource use to pull new connections to set.
     */
    public void setDataSource(DataSource dataSource) {
        this.dataSource = dataSource;
    }    

}
