/***********************************************************
 * $Id: JDBCHelper.java 1128 2016-02-21 20:36:47Z util $
 * 
 * SQL/DAO utilities of clazzes.org
 * http://www.clazzes.org
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 * 
 ***********************************************************/

package org.clazzes.util.sql.helper;

import java.sql.Connection;
import java.sql.PreparedStatement;
import java.sql.ResultSet;
import java.sql.SQLException;
import java.sql.Timestamp;
import java.sql.Types;

/**
 * <p>Static helper functions for Number object binding/fetching
 * from JDBC result sets and prepared statements.</p>
 * 
 * <p>These functions greatly ease the life
 * of programmers when working with
 * {@link Long}, {@link Integer},... properties, because the
 * plain JDBC API is very tedious to use.</p>
 */
public abstract class JDBCHelper {

    /**
     * Format the transaction isolation level constants as a string.
     * 
     * @param isolationLevel A transaction isolation level constant.
     * @return A string representation.
     * 
     * @see Connection#TRANSACTION_NONE
     * @see Connection#TRANSACTION_READ_UNCOMMITTED
     * @see Connection#TRANSACTION_READ_COMMITTED
     * @see Connection#TRANSACTION_REPEATABLE_READ
     * @see Connection#TRANSACTION_SERIALIZABLE
     */
    public static final String formatIsolationLevel(int isolationLevel) {
        
        switch (isolationLevel) {
        
        case Connection.TRANSACTION_NONE:
            return "TRANSACTION_NONE";
        
        case Connection.TRANSACTION_READ_UNCOMMITTED:
            return "TRANSACTION_READ_UNCOMMITTED";

        case Connection.TRANSACTION_READ_COMMITTED:
            return "TRANSACTION_READ_COMMITTED";
        
        case Connection.TRANSACTION_REPEATABLE_READ:
            return "TRANSACTION_REPEATABLE_READ";
        
        case Connection.TRANSACTION_SERIALIZABLE:
            return "TRANSACTION_SERIALIZABLE";

        default:
            return "TRANSACTION_ISOLATION_UNKNOWN" + isolationLevel;
        }
    }
    
    
	/**
	 * @param rs A result set.
	 * @param columnIndex The column index of the column to get;
	 *                    the first column has the index 1.
	 * @return A Boolean object or <code>null</code>, if
	 *              {@link ResultSet#wasNull()} returned <code>true</code>.
	 * @throws SQLException Upon errors.
	 */
	public static Boolean getBoolean(ResultSet rs, int columnIndex) throws SQLException {
		
		boolean b = rs.getBoolean(columnIndex);
		return rs.wasNull() ? null : Boolean.valueOf(b);
	}
	
	/**
	 * @param rs A result set.
	 * @param columnLabel The column label of the column to get.
	 * @return A Boolean object or <code>null</code>, if
	 *              {@link ResultSet#wasNull()} returned <code>true</code>.
	 * @throws SQLException Upon errors.
	 */
	public static Boolean getBoolean(ResultSet rs, String columnLabel) throws SQLException {
		
		boolean b = rs.getBoolean(columnLabel);
		return rs.wasNull() ? null : Boolean.valueOf(b);
	}

	/**
	 * @param rs A result set.
	 * @param columnIndex The column index of the column to get;
	 *                    the first column has the index 1.
	 * @return A Byte object or <code>null</code>, if
	 *              {@link ResultSet#wasNull()} returned <code>true</code>.
	 * @throws SQLException Upon errors.
	 */
	public static Byte getByte(ResultSet rs, int columnIndex) throws SQLException {
		
		byte b = rs.getByte(columnIndex);
		return rs.wasNull() ? null : Byte.valueOf(b);
	}
	
	/**
	 * @param rs A result set.
	 * @param columnLabel The column label of the column to get.
	 * @return A Byte object or <code>null</code>, if
	 *              {@link ResultSet#wasNull()} returned <code>true</code>.
	 * @throws SQLException Upon errors.
	 */
	public static Byte getByte(ResultSet rs, String columnLabel) throws SQLException {
		
		byte b = rs.getByte(columnLabel);
		return rs.wasNull() ? null : Byte.valueOf(b);
	}

	/**
	 * @param rs A result set.
	 * @param columnIndex The column index of the column to get;
	 *                    the first column has the index 1.
	 * @return A Short object or <code>null</code>, if
	 *              {@link ResultSet#wasNull()} returned <code>true</code>.
	 * @throws SQLException Upon errors.
	 */
	public static Short getShort(ResultSet rs, int columnIndex) throws SQLException {
		
		short s = rs.getShort(columnIndex);
		return rs.wasNull() ? null : Short.valueOf(s);
	}
	
	/**
	 * @param rs A result set.
	 * @param columnLabel The column label of the column to get.
	 * @return A Short object or <code>null</code>, if
	 *              {@link ResultSet#wasNull()} returned <code>true</code>.
	 * @throws SQLException Upon errors.
	 */
	public static Short getShort(ResultSet rs, String columnLabel) throws SQLException {
		
		short s = rs.getShort(columnLabel);
		return rs.wasNull() ? null : Short.valueOf(s);
	}

	/**
	 * @param rs A result set.
	 * @param columnIndex The column index of the column to get;
	 *                    the first column has the index 1.
	 * @return An Integer object or <code>null</code>, if
	 *              {@link ResultSet#wasNull()} returned <code>true</code>.
	 * @throws SQLException Upon errors.
	 */
	public static Integer getInt(ResultSet rs, int columnIndex) throws SQLException {
		
		int i = rs.getInt(columnIndex);
		return rs.wasNull() ? null : Integer.valueOf(i);
	}
	
	/**
	 * @param rs A result set.
	 * @param columnLabel The column label of the column to get.
	 * @return An Integer object or <code>null</code>, if
	 *              {@link ResultSet#wasNull()} returned <code>true</code>.
	 * @throws SQLException Upon errors.
	 */
	public static Integer getInt(ResultSet rs, String columnLabel) throws SQLException {
		
		int i = rs.getInt(columnLabel);
		return rs.wasNull() ? null : Integer.valueOf(i);
	}

	/**
	 * @param rs A result set.
	 * @param columnIndex The column index of the column to get;
	 *                    the first column has the index 1.
	 * @return A Long object or <code>null</code>, if
	 *              {@link ResultSet#wasNull()} returned <code>true</code>.
	 * @throws SQLException Upon errors.
	 */
	public static Long getLong(ResultSet rs, int columnIndex) throws SQLException {
		
		long l = rs.getLong(columnIndex);
		return rs.wasNull() ? null : Long.valueOf(l);
	}
	
	/**
	 * @param rs A result set.
	 * @param columnLabel The column label of the column to get.
	 * @return A Long object or <code>null</code>, if
	 *              {@link ResultSet#wasNull()} returned <code>true</code>.
	 * @throws SQLException Upon errors.
	 */
	public static Long getLong(ResultSet rs, String columnLabel) throws SQLException {
		
		long l = rs.getLong(columnLabel);
		return rs.wasNull() ? null : Long.valueOf(l);
	}

	/**
	 * @param rs A result set.
	 * @param columnIndex The column index of the column to get;
	 *                    the first column has the index 1.
	 * @return A Float object or <code>null</code>, if
	 *              {@link ResultSet#wasNull()} returned <code>true</code>.
	 * @throws SQLException Upon errors.
	 */
	public static Float getFloat(ResultSet rs, int columnIndex) throws SQLException {
		
		float f = rs.getFloat(columnIndex);
		return rs.wasNull() ? null : Float.valueOf(f);
	}
	
	/**
	 * @param rs A result set.
	 * @param columnLabel The column label of the column to get.
	 * @return A Float object or <code>null</code>, if
	 *              {@link ResultSet#wasNull()} returned <code>true</code>.
	 * @throws SQLException Upon errors.
	 */
	public static Float getFloat(ResultSet rs, String columnLabel) throws SQLException {
		
		float f = rs.getFloat(columnLabel);
		return rs.wasNull() ? null : Float.valueOf(f);
	}

	/**
	 * @param rs A result set.
	 * @param columnIndex The column index of the column to get;
	 *                    the first column has the index 1.
	 * @return A Double object or <code>null</code>, if
	 *              {@link ResultSet#wasNull()} returned <code>true</code>.
	 * @throws SQLException Upon errors.
	 */
	public static Double getDouble(ResultSet rs, int columnIndex) throws SQLException {
		
		double d = rs.getDouble(columnIndex);
		return rs.wasNull() ? null : Double.valueOf(d);
	}
	
	/**
	 * @param rs A result set.
	 * @param columnLabel The column label of the column to get.
	 * @return A Double object or <code>null</code>, if
	 *              {@link ResultSet#wasNull()} returned <code>true</code>.
	 * @throws SQLException Upon errors.
	 */
	public static Double getDouble(ResultSet rs, String columnLabel) throws SQLException {
		
		double d = rs.getDouble(columnLabel);
		return rs.wasNull() ? null : Double.valueOf(d);
	}
	
    /** Helper function for the case, that utc seconds are stored in Java code in an attribute of type Double,
     *  and mapped to a database column of type e.g. date, datetime or timestamp.  Reads a java.sql.Date from
     *  the given ResultSet, and transforms it into utcSeconds stored in a Double object.
     *  WARNING: The result depends on the time zone of the JVM.
     *  @param resultSet the resultSet
     *  @param columnIndex the column index of the column to get; the first column has the index 1.
     *  @return the utcSeconds read from the ResultSet, null if the database entry was null
     */	
	public static Double getUtcSeconds(ResultSet resultSet, int columnIndex) throws SQLException {
        Timestamp timeStamp = resultSet.getTimestamp(columnIndex);
        return timeStamp == null ? null : new Double((double)timeStamp.getTime() / 1000); 
    }	

	/**
	 * Call <code>ps.setBoolean(parameterIndex,b.booleanValue())</code> or
	 * <code>ps.setNull(parameterIndex,Types.BOOLEAN)</code> when <code>b</code>
	 * is <code>null</code>.
	 * 
	 * @param ps A prepared statement.
	 * @param parameterIndex The index of the placeholder in the statement to fill;
	 *                the first placeholder has the index 1.
	 * @param b A Boolean object.
	 * @throws SQLException Upon errors.
	 */
	public static void setBoolean(PreparedStatement ps, int parameterIndex, Boolean b) throws SQLException {
		
		if (b == null) {
			ps.setNull(parameterIndex,Types.BOOLEAN);
		}
		else {
			ps.setBoolean(parameterIndex,b.booleanValue());
		}
	}
	
	/**
	 * Call <code>ps.setByte(parameterIndex,b.byteValue())</code> or
	 * <code>ps.setNull(parameterIndex,Types.TINYINT)</code> when <code>b</code>
	 * is <code>null</code>.
	 * 
	 * @param ps A prepared statement.
	 * @param parameterIndex The index of the placeholder in the statement to fill;
	 *                the first placeholder has the index 1.
	 * @param b A Byte object.
	 * @throws SQLException Upon errors.
	 */
	public static void setByte(PreparedStatement ps, int parameterIndex, Number b) throws SQLException {
		
		if (b == null) {
			ps.setNull(parameterIndex,Types.TINYINT);
		}
		else {
			ps.setByte(parameterIndex,b.byteValue());
		}
	}
	
	/**
	 * Call <code>ps.setShort(parameterIndex,s.shortValue())</code> or
	 * <code>ps.setNull(parameterIndex,Types.SMALLINT)</code> when <code>b</code>
	 * is <code>null</code>.
	 * 
	 * @param ps A prepared statement.
	 * @param parameterIndex The index of the placeholder in the statement to fill;
	 *                the first placeholder has the index 1.
	 * @param s A Short object.
	 * @throws SQLException Upon errors.
	 */
	public static void setShort(PreparedStatement ps, int parameterIndex, Number s) throws SQLException {
		
		if (s == null) {
			ps.setNull(parameterIndex,Types.SMALLINT);
		}
		else {
			ps.setShort(parameterIndex,s.shortValue());
		}
	}
	
	/**
	 * Call <code>ps.setInt(parameterIndex,i.intValue())</code> or
	 * <code>ps.setNull(parameterIndex,Types.INTEGER)</code> when <code>b</code>
	 * is <code>null</code>.
	 * 
	 * @param ps A prepared statement.
	 * @param parameterIndex The index of the placeholder in the statement to fill;
	 *                the first placeholder has the index 1.
	 * @param i An Integer object.
	 * @throws SQLException Upon errors.
	 */
	public static void setInt(PreparedStatement ps, int parameterIndex, Number i) throws SQLException {
		
		if (i == null) {
			ps.setNull(parameterIndex,Types.INTEGER);
		}
		else {
			ps.setInt(parameterIndex,i.intValue());
		}
	}
	
	/**
	 * Call <code>ps.setLong(parameterIndex,l.longValue())</code> or
	 * <code>ps.setNull(parameterIndex,Types.BIGINT)</code> when <code>b</code>
	 * is <code>null</code>.
	 * 
	 * @param ps A prepared statement.
	 * @param parameterIndex The index of the placeholder in the statement to fill;
	 *                the first placeholder has the index 1.
	 * @param l A Long object.
	 * @throws SQLException Upon errors.
	 */
	public static void setLong(PreparedStatement ps, int parameterIndex, Number l) throws SQLException {
		
		if (l == null) {
			ps.setNull(parameterIndex,Types.BIGINT);
		}
		else {
			ps.setLong(parameterIndex,l.longValue());
		}
	}
	
	/**
	 * Call <code>ps.setFloat(parameterIndex,f.floatValue())</code> or
	 * <code>ps.setNull(parameterIndex,Types.FLOAT)</code> when <code>b</code>
	 * is <code>null</code>.
	 * 
	 * @param ps A prepared statement.
	 * @param parameterIndex The index of the placeholder in the statement to fill;
	 *                the first placeholder has the index 1.
	 * @param f A Float object.
	 * @throws SQLException Upon errors.
	 */
	public static void setFloat(PreparedStatement ps, int parameterIndex, Number f) throws SQLException {
		
		if (f == null) {
			ps.setNull(parameterIndex,Types.FLOAT);
		}
		else {
			ps.setFloat(parameterIndex,f.floatValue());
		}
	}
	
	/**
	 * Call <code>ps.setDouble(parameterIndex,d.doubleValue())</code> or
	 * <code>ps.setNull(parameterIndex,Types.DOUBLE)</code> when <code>b</code>
	 * is <code>null</code>.
	 * 
	 * @param ps A prepared statement.
	 * @param parameterIndex The index of the placeholder in the statement to fill;
	 *                the first placeholder has the index 1.
	 * @param d A Double object.
	 * @throws SQLException Upon errors.
	 */
	public static void setDouble(PreparedStatement ps, int parameterIndex, Number d) throws SQLException {
		
		if (d == null) {
			ps.setNull(parameterIndex,Types.DOUBLE);
		}
		else {
			ps.setDouble(parameterIndex,d.doubleValue());
		}
	}
	
	/** Helper function for the case, that utc seconds are stored in Java code in an attribute of type Double,
	 *  and mapped to a database column of type e.g. date, datetime or timestamp.  Transforms the given
	 *  utcSeconds into a java.sql.Date and calls the corresponding setter on the given statement.
     *  WARNING: The result depends on the time zone of the JVM.
	 * 
	 * @param statement statement
	 * @param parameterIndex The index of the placeholder in the statement to fill;
     *                the first placeholder has the index 1.
	 * @param utcSeconds utc seconds as a Double object, can be null
	 * @throws SQLException upon errors
	 */
	public static void setUtcSeconds(PreparedStatement statement, int parameterIndex, Double utcSeconds) throws SQLException {
	    if (utcSeconds == null) {
	        statement.setNull(parameterIndex, Types.DATE);
	    } else {
	        long utcMillis = (long)(utcSeconds.doubleValue() * 1000);
	        Timestamp timeStamp = new Timestamp(utcMillis);
	        statement.setTimestamp(parameterIndex, timeStamp);
	    }
	}

	/**
	 * Set an enum value as an integer parameter to a prepared statement
	 * using the ordinal of an enum value. If the enum value is <code>null</code>,
	 * <code>statement.setNull(pos,Types.INTEGER)</code> is performed.
	 * 
	 * @param statement A prepared statement.
	 * @param pos The parameterIndex
	 * @param val The enum value to set.
	 * @throws SQLException Upon DB errors.
	 */
	public static <T extends Enum<T>> void setEnum(PreparedStatement statement, int pos, T val) throws SQLException {
		
		if (val == null) {
			statement.setNull(pos,Types.INTEGER);
		}
		else {
			statement.setInt(pos,val.ordinal());
		}
	}

	/**
	 * Fetch an enum value from an integer column of a result set interpreted
	 * as the ordinal of the given enum class.
	 * 
	 * @param rs The result set to get the enum from.
	 * @param pos The index of the value to get.
	 * @param clazz The class of the requested enum value.
	 * @return An enum or <code>null</code>, if the column of the result set was null.
	 * @throws SQLException Upon DB errors.
	 */
	public static <T extends Enum<T>> T getEnum(ResultSet rs, int pos, Class<T> clazz) throws SQLException {
	
		int o = rs.getInt(pos);
		
		if (rs.wasNull()) {
			return null;
		}
		else {
			return clazz.getEnumConstants()[o];
		}	
	}

}
